// -*- mode: c++; indent-tabs-mode: nil -*-
//! @file WSDL.qm WSDL: Web Services Description Language: http://www.w3.org/TR/wsdl

/*  WSDL.qm Copyright (C) 2012 - 2014 David Nichols

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

// make sure we have the required qore version

// requires XML functionality

// need mime definitions



/*  Version History
    * 2014-01-14 v0.3.5: David Nichols <david@qore.org>: fixed msg serialization issues
    * 2012-05-31 v0.3.4: David Nichols <david@qore.org>: updated to a user module
    *            v0.3.3: added initial support for the anyAttribute element of complexType
                         added initial support for SOAP header processing
                         added initial support for multiple portType and bindings in a WSDL

    *            v0.3.2: fixed bugs de/serializing negative values for "int" and "short"
    *            v0.3.1: improved XSD imports and namespace handling

    *            v0.3.0: implemented WSDLLib class of helper functions
                         implemented support for xsd import statements in WSDLs

    *            v0.2.0: use parseXMLAsData() instead of parseXML()
                         implemented initial simpleType support
                         fixed xsd:date serialization and deserialization


    WSDL classes
    provides some minimal WSDL and XSD support for SOAP messaging used by the SoapClient class and the SoapHandler

    not complete, needs namespace verification, improved XSD support, element groups, etc
*/

/** @mainpage WSDL Module
    
    @section wsdlintro Introduction to the WSDL Module

    The %WSDL module provides functionality for <a href="http://en.wikipedia.org/wiki/SOAP">SOAP</a>.

    The main purpose of this module is to provide the infrastructure for developing SOAP client and server services.

    The main components of this module are as follows:
    - @ref WSDL::WebService "WebService": represents a WSDL file; an easy way to get an object of this type from a URL is by calling WSDL::WSDLLib::getWSDL()
    - @ref WSDL::WSDLLib "WSDLLib": a class of static functions providing helper functions for the module
    - @ref WSDL::WSOperation "WSOperation": a class representing a single operation from a web service

    See the following for modules using the %WSDL module:
    - <a href="../../SoapClient/html/index.html">SoapClient user module</a>
    - <a href="../../SoapHandler/html/index.html">SoapHandler user module</a>
*/

//! main WSDL namespace
namespace WSDL {
    //! this WSDL implementation version
    const version     = "0.3.5";
    
    const XET_ALL      = "ALL";
    const XET_CHOICE   = "CHOICE";
    const XET_SEQUENCE = "SEQUENCE";

    //! SOAP 1.1 envelope URI
    const SOAP_11_ENV  = "http://schemas.xmlsoap.org/soap/envelope/";
    //! SOAP 1.2 envelope URI
    const SOAP_12_ENV  = "http://www.w3.org/2003/05/soap-envelope";

    //! SOAP 1.2 namespace URI
    const SOAP_12_NS   = "http://schemas.xmlsoap.org/wsdl/soap12/";

    //! XSD namespace URI
    const XSD_NS       = "http://www.w3.org/2001/XMLSchema";
    //! XSI namespace URI
    const XSI_NS       = "http://www.w3.org/2001/XMLSchema-instance";

    //! HTTP namespace URI
    const HTTP_NS      = "http://schemas.xmlsoap.org/wsdl/http/";
    //! MIME namespace URI
    const MIME_NS      = "http://schemas.xmlsoap.org/wsdl/mime/";

    //! soap 1.1 envelope namespaces
    const ENVELOPE_11_NS = 
	( "soapenv:Envelope" :
	  ( "^attributes^" : 
	    ( "xmlns:soapenv" : SOAP_11_ENV,
	      "xmlns:xsd"     : XSD_NS,
	      "xmlns:xsi"     : XSI_NS ) ) );

    //! soap 1.2 envelope namespaces
    const ENVELOPE_12_NS = 
	( "soapenv:Envelope" :
	  ( "^attributes^" : 
	    ( "xmlns:soapenv" : SOAP_12_ENV,
	      "xmlns:xsd"     : XSD_NS,
	      "xmlns:xsi"     : XSI_NS ) ) );

    //! soap encoding URI
    const SOAP_ENCODING = "http://schemas.xmlsoap.org/soap/encoding/";

    //! mapping from Qore types to xsd types for xsd type "anyType"
    const any_type_map = 
	( Type::String      : "string",
	  Type::Int         : "long",
	  Type::Boolean     : "boolean",
	  Type::Date        : "dateTime",
	  Type::Float       : "decimal",
	  Type::NothingType : "string",
	  Type::NullType    : "string",
	  Type::Binary      : "base64Binary" );

    // error codes
    const SOAP_SERIALIZATION_ERROR = "SOAP-SERIALIZATION-ERROR";
    const SOAP_DESERIALIZATION_ERROR = "SOAP-DESERIALIZATION-ERROR";
    const WSDL_ERROR = "WSDL-ERROR";

    //! SOAP transport URL
    const SOAP_TRANSPORT_HTTP = "http://schemas.xmlsoap.org/soap/http";

    //! range of "short" values (16 bits)
    const RANGE_SHORT = (-32768, 32767);

    //! range of "int" values (32 bits)
    const RANGE_INT = (-2147483648, 2147483647);
};

//! contains helper methods for retrieving WSDLs from a URL
class WSDL::WSDLLib {

public:
    //! Mime types recognized as SOAP messages
    const SoapMimeTypes = (MimeTypeSoapXml, MimeTypeXml, MimeTypeXmlApp);

    //! retrieves a local file and returns the file's contents as a string
/** called by WSDLLib::getFileFromURL() in case the scheme is "file"
    */
      static string getFile(string fn, bool as_string = True);

    //! retrieves a file from a URL with HTTP and returns the file's contents as a string
/** called by WSDLLib::getFileFromURL() in case the scheme is "http"
    */
      static string getHTTP(string url, __7_ string path, __7_ HTTPClient hc, __7_ hash headers, bool as_string = True);

    //! retrieves a file from a URL with the FTP protocol and returns the file's contents as binary or string data
/** called by WSDLLib::getFileFromURL() in case the scheme is "ftp"
    */
      static data getFTP(string url, string path, bool as_string = True);

    //! retrieves a file from a URL
    static data getFileFromURL(string url, string def_protocol = "file", __7_ HTTPClient http_client, __7_ hash http_headers, bool as_string = True);


    //! retrieves a file from an already-parsed URL
    static data getFileFromURL(string url, hash u, string def_protocol = "file", __7_ HTTPClient http_client, __7_ hash http_headers, bool as_string = False);


    //! returns the argument
      static WebService getWSDL(WebService wsdl);

    //! returns a WSDL string form a file name, optional HTTPClient object and optional header hash
      static string getWSDL(string wsdl, __7_ HTTPClient http_client, __7_ hash http_headers);

    //! takes a hash representation of a SOAP message and parses it to a Qore data structure; handles multipart messages, checks the content-type, and handles hrefs in the message
      static hash parseSOAPMessage(hash msg);


private:
      static checkSOAPContentType(string ct);
public:



private:
      static processHref(reference xmldata, string hr, hash parts);
public:

    

private:
      static substHref(reference xmldata, hash parts);
public:

};

//! base class with helper methods for XSD data processing
class WSDL::XSDBase {

public:
    /*
    static private hash doType(string $t) {
	#printf("DEBUG: XSDBase::doType(%y)\n", $t);
	my (__7_ string $ns, __7_ string $type) = $t =~ x/(\w+):(\w+)/;
	return !exists $type ? ("val": $t) : ("ns": $ns, "val": $type);
    }
    */

      static removeNS(reference v);

      static removeNS2(reference v);
};

//! base class for XSD data classes
class WSDL::XSDData : public WSDL::XSDBase {

public:
    any getValue(__7_ hash mrh, any val);

};

//! base class for XSD classes with a "name" attribute
class WSDL::XSDNamedData : public WSDL::XSDData {

public:
    public :
        // name of object
        string name;
        //! input namespace prefix (if any given)
        __7_ string ns;

public:

    constructor(string name, string ns);


    constructor(reference e);


    string getName();


    __7_ string getInputNamespacePrefix();

};

class WSDL::XSDAbstractType : public WSDL::XSDNamedData {

public:
    public :
        //! reference to namespaces
        Namespaces nsc;
        
        //! namespace output prefix
        string ons;

public:

    constructor(reference e, Namespaces nsc);


    constructor(string name, string ns, Namespaces nsc);


    private resolveNamespace();


    checkExtends(XSDAbstractType t, string ename);


    string getNameWithNS();


    bool required();


    string getOutputNamespacePrefix();


    abstract any serialize(any val, __7_ softbool omit_type, __7_ softbool omit_ns);
    abstract any deserialize(string en, __7_ hash types, __7_ hash mrh, any val);
};

//! class for XSD base types
class WSDL::XSDBaseType : public WSDL::XSDAbstractType {

public:
    public :

public:

    constructor(string t, Namespaces nsc);


    any serialize(any val, __7_ softbool omit_type, __7_ softbool omit_ns);


    any deserialize(string en, __7_ hash types, __7_ hash mrh, any val);

};

//! class for XSD array types; currently only supports "binary"; used, for example with HTTP MultiPart messages
class WSDL::XSDArrayType : public WSDL::XSDAbstractType {

public:
    constructor(string t, Namespaces nsc);


    any serialize(any val, __7_ softbool omit_type, __7_ softbool omit_ns);


    any deserialize(string en, __7_ hash types, __7_ hash mrh, any val);

};

//! XSD element class
class WSDL::XSDElement : public WSDL::XSDNamedData {

public:
    public :
        any type;
        int minOccurs = 1;
        int maxOccurs = 1;
        bool nillable = False;

public:

    constructor(hash e, Namespaces nsc, __7_ XSDAbstractType type, XSDLateResolverHelper unresolved);


    bool required();

    
    any serialize(any h, __7_ softbool omit_type, __7_ softbool omit_ns, string key, string typename, reference ons);


    any serializeAs(XSDAbstractType type, any h, __7_ softbool omit_type, __7_ softbool omit_ns, string key, string typename, reference ons);


    private any serializeAsIntern(XSDAbstractType type, any h, __7_ softbool omit_type, __7_ softbool omit_ns, string key, string typename, reference ons);


    any deserialize(__7_ hash types, __7_ hash mrh, any val);

};

//! XSD simple type class
class WSDL::XSDSimpleType : public WSDL::XSDAbstractType {

public:
    public :
        hash enum;
        any type;

public:

    constructor(hash st, Namespaces nsc, XSDLateResolverHelper unresolved);


    any serialize(any val, __7_ softbool omit_type, __7_ softbool omit_ns);


    any deserialize(string en, __7_ hash types, __7_ hash mrh, any val);

};

//! XSD complex type class
class WSDL::XSDComplexType : public WSDL::XSDAbstractType {

public:
    public :
        any array;
        any restriction;
        __7_ string extension;
        any hash_type;
        __7_ hash elements;

	bool anyAttribute = False;

        bool required = False;

        // any annotated documentation
        __7_ string documentation;

public:

    constructor(any ct, Namespaces nsc, XSDLateResolverHelper unresolved);


    checkExtends(XSDAbstractType t, string ename);


    private parseData(any d, XSDLateResolverHelper unresolved);


    bool required();


    private __7_ hash parseElements(any el, XSDLateResolverHelper unresolved);


    private __7_ hash serializeElement(string key, any h, __7_ softbool omit_type, __7_ softbool omit_ns);


    __7_ hash serialize(any h, __7_ softbool omit_type, __7_ softbool omit_ns);

    
    __7_ hash deserialize(string en, __7_ hash types, __7_ hash mrh, any oval);

};

//! web service operation class
class WSDL::WSOperation : public WSDL::XSDNamedData {

public:
    public :
        WSMessage input;
        WSMessage output;
        __7_ string input_name;
        __7_ string output_name;
	any types;

        Namespaces nsc;

	bool usedocns = False;
        __7_ string soapAction;
        __7_ string request_name;
        any in;
        any out;
        bool docstyle = False;
	
	// info about soap header requirements
	hash iheader;
	hash oheader;

public:

    constructor(any p, __7_ hash types, Namespaces nsc, __7_ hash messages, bool usedocns = False);


    setDocStyle(reference idocmap);
	

    setTopLevelRequestElement(string name);


    string getTopLevelRequestName();


    //! serializes a request to an XML string for the operation
/** @param h the request to serialize
	@param header optional soap header info to serialize if required (ex: authorization info)
        @param enc the optional encoding to use; if this argument is not present, then the default encoding will be used

	@return a hash with keys:
	- \c body: XML string in the SOAP request format
	- \c hdr: hash of HTTP headers
	*/
    hash serializeRequest(any h, __7_ hash header, __7_ string enc, __7_ hash nsh);


    //! serializes a SOAP response to an XML string for the operation
/** @param h the response to serialize
        @param enc the optional encoding to use; if this argument is not present, then the default encoding will be used
        @return a hash with keys:
        - \c body: XML string in the SOAP request format
        - \c hdr: hash of HTTP headers
    */
    hash serializeResponse(any h, __7_ hash header, __7_ string enc, __7_ hash nsh);


    private list processMultiRef(hash body);


    //! parses a hash representing a parsed XML request (parsed with parseXMLAsData()) for the operation and returns the corresponding Qore data structure
/** @param o the parsed XML request (parsed with parseXMLAsData()) for the operation
    @return the Qore data structure corresponding to the request data
    */
    any deserializeRequest(hash o);


    //! parses a hash representing a parsed XML response (parsed with parseXMLAsData()) for the operation and returns the corresponding Qore data structure
/** @param o the parsed XML response (parsed with parseXMLAsData()) for the operation
    @return the Qore data structure corresponding to the response data
    */
    any deserializeResponse(hash o);


    private hash processNSValue(hash h);


    //! returns True if the operation is a SOAP 1.2 operation
/** @return True if the operation is a SOAP 1.2 operation
    */
    bool isSoap12();


    //! returns the target namespace for the operation
/** @return the target namespace for the operation
    */
    string getTargetNS();


    setOutputMultipart(any v);


    private parsePart(reference msg, any part);


    addOutputPart(any part);


    setInputMultipart(any v);


    addInputPart(any part);


    setInputHeader(string part, WSMessage msg, bool encoded);

};

//! web service message class
class WSDL::WSMessage : public WSDL::XSDNamedData {

public:
    public :
        hash args;
        // part map; maps element names to part names
        hash pmap;
        bool encoded = False;

        // keep a reference to Namespaces
        Namespaces nsc;

public:

    constructor(hash m, __7_ hash element_map, Namespaces nsc);


    hash serialize(any msginfo, any mpm, string name, any h);


    hash serializeDocument(__7_ string k, any msginfo, any mpm, any h, bool force_ns);


    any deserialize(__7_ hash types, __7_ hash mrh, hash val);


    any deserializeDocument(__7_ hash types, __7_ hash mrh, any val);

};

// private helper class for lazy name resolution
class WSDL::XSDLateResolverHelper {

public:
private:
 list l = (); 
public:

    constructor();


    add(any v);


    list getList();

};

//! class for WSDL bindings
class WSDL::Binding : public WSDL::XSDNamedData {

public:
    public :
	bool docStyle = False;

public:

    constructor(hash data, Namespaces nsc, hash opmap, reference idocmap, __7_ hash messages);

};

//! namespace container class
class WSDL::Namespaces {

public:
    public :

public:

    private :
        //! maps output namespace prefixes to namespace URIs
        hash ns = (
            "xsd": XSD_NS,
            );

        //! maps namespace URIs to output namespace prefixes
        hash nsr = (
            XSD_NS: "xsd",
            );

        //! hash for valid XSD namespaces, maps input namespace prefixes -> True if it refers to an XSD schema
        hash xml_schema;

        //! hash mapping input namespace prefixes to namespaces URIs
        hash imap = (
            "xsd": XSD_NS,
            );

        //! integer providing a sequence for output namespace prefixes
        int nsn;

        //! if True then using SOAP 1.2
        bool soap12 = False;

        //! current target namespace
        __7_ string target_ns;

        //! target namespace stack;
        list nss = ();

public:

    //! creates the object with the wsdl definitions attribute hash
    constructor(hash nsh, __7_ Namespaces nsc);


    //! merges namespaces when parsing imported schemas
    merge(Namespaces nsc);


    //! returns the namespace URI for the given output namespace prefix
    string getOutputNamespaceUri(string nsp);


    //! returns a hash of namespace attributes for outgoing SOAP messages
    __7_ hash getOutputNamespaceHash(__7_ hash nsh);


    //! returns the primary target namespace Uri
    __7_ string getTargetNamespaceUri();


    //! pushes the previous target namespace URI on the stack when parsing schemas and sets the current target namespace URI to the current value
    pushTargetNamespace(string ns);


    //! restores any previous target namespace URI from the stack to the current target namespace URI
    popTargetNamespace();


    //! looks up and registers a namespace if necessary, returns the namespace's prefix
    string getOutputNamespacePrefix(string ns);


    //! registers a namespace internally
    private string registerNamespaceIntern(string ns);


    //! returns a hash of namespace prefixes to namespaces URIs actually used
    __7_ hash getReferencedNamespaceMap();


    //! returns True if using SOAP 1.2, False if not
    bool isSoap12();


    //! returns True if if the input namespace prefix refers to the XSD namespace URI
    __7_ bool isSchema(string ns);


    //! returns the input namespace URI from the input namespace prefix
    string getInputNamespaceUri(string nsa);


    any doType(string t);


    //! returns the output namespace prefix for the given input namespace prefix
    string translateOutputNamespacePrefix(__7_ string nsa);

};

//! main class representing a parsed WSDL file
/** This is the main class for handling SOAP communication and is based on a WSDL file
*/
class WSDL::WebService : public WSDL::XSDBase {

public:
    public :
        //! the WSDL string
        string wsdl;

        //! namespace container
        Namespaces nsc;

        //! namespace info
        //hash ns;

        //bool soap12 = False;
        hash base_type;
        hash services;
        list wsdl_services = ();
        hash idocmap;
        hash opmap;
        hash binding;
        hash element_map;
        hash messages;
        hash types;
        bool usedocns = False;
        hash portType;
        __7_ code try_import;

public:

    //! creates the WebService object
/** @param str the XML string representing the WSDL
	@param opts an optional hash of options with the following possible keys:
        - \c try_import a call reference or closure to be passed a string name for XSD imports without any scheme to retrieve the data, must take a string argument (the resource name) and return a string (the resource data)
	- \c http_client a HTTPClient object for retrieving data from import commands
	- \c http_headers a hash of optional HTTP header info to use when retrieving data from import commands
    */
    constructor(string str, __7_ hash opts);


    hash getType(string name, any v);


    //! returns a map of top-level element names to WSOperation objects
/** @return a map of top-level element names to WSOperation objects
    */
    __7_ hash getOperationMap(any name);


    /*
    private getNSPrefixes(__7_ hash $a) {
	#printf("DEBUG: getNSPrefixes() %n\n", $a);
	foreach my string $k in ($a.keyIterator()) {
	    my any $ns = ($k =~ x/xmlns:(\w+)/)[0];
	    
	    if (!exists $ns)
		continue;
	    ns.map.$ns = $a.$k;

	    if ($a.$k == XSD_NS)
		ns.xml_schema.$ns = True;

	    if ($a.$k == SOAP_12_NS)
		soap12 = True;
	}
	# "default" has to be quoted because it's a reserved word
	if (exists $a.xmlns)
	    ns."default" = $a.xmlns;
	ns.target = $a.targetNamespace;
    }
    */

    private XSDBaseType getBaseType(any t);


    private XSDAbstractType resolveType(hash v);


    // parse XSD schema types
    private parseTypes(__7_ hash data, any http_client, any http_headers);


    private parseMessages(__7_ softlist message);


    private parseService(any svc);


    private parsePortType(any data);


    private parseBinding(any bindings);


    //! returns True if the WSDL describes a SOAP 1.2 service
/** @return True if the WSDL describes a SOAP 1.2 service
    */
    bool isSoap12();


    //! returns the XML string for the WSDL
/** @return the XML string for the WSDL
    */
    string getWSDL();

};
